from scbamtools.contrib import __version__, __license__, __author__, __email__
import numpy as np
from collections import defaultdict
import mrfifo as mf
import scbamtools.util as util
import scbamtools.config as config


def parse_args():
    parser = util.make_minimal_parser(
        prog="trim.py",
        description="trim adapters from a BAM file using cutadapt",
    )
    parser.add_argument(
        "--config", default="", help="path to custom config-file (default='')"
    )
    parser.add_argument(
        "bam_in",
        help="bam input (default=stdin)",
        default="/dev/stdin",
        # nargs="+",
    )
    parser.add_argument(
        "--bam-out",
        help="bam output (default=stdout)",
        default="/dev/stdout",
    )
    parser.add_argument(
        "--bam-out-mode",
        help="bam output mode (default=b)",
        default="b",
    )
    parser.add_argument(
        "--flavor",
        help="name of the adapter flavor used to retrieve sequences and parameters from the config.yaml",
        default="default",
    )
    parser.add_argument(
        "--skim",
        help="skim through the BAM by investigating only every <skim>-th record (default=1 off)",
        default=1,
        type=int,
    )
    parser.add_argument(
        "--phred-base",
        help="phred score base used for qual trimming (default=33)",
        type=int,
        default=33,
    )
    parser.add_argument(
        "--threads-read",
        help="number of threads for reading bam_in (default=2)",
        type=int,
        default=2,
    )
    parser.add_argument(
        "--threads-write",
        help="number of threads for writing bam_out (default=4)",
        type=int,
        default=4,
    )
    parser.add_argument(
        "--threads-work",
        help="number of worker threads for actual trimming (default=8)",
        type=int,
        default=8,
    )
    parser.add_argument(
        "--stats-out",
        help="write tab-separated table with trimming results here",
        default="",
    )
    args = parser.parse_args()
    return config.load(args.config, args=vars(args))


class QualityTrim:
    def __init__(self, min_base_qual=20):
        self.min_base_qual = min_base_qual
        self.name = "Q"

    def match_to(self, seq, qual):
        from cutadapt.qualtrim import quality_trim_index

        _, end = quality_trim_index(qual, 0, self.min_base_qual)
        # we use the cutadapt function here (which implements BWA's logic).
        # but only for the 3' end.
        n_trimmed = len(qual) - end
        return n_trimmed


class AdapterTrim:
    def __init__(self, name="na", where="right", seq=None, **kw):
        import cutadapt.adapters

        self.name = name
        self.where = where
        if where == "left":
            self.adapter = cutadapt.adapters.NonInternalFrontAdapter(
                seq, name=name, **kw
            )
        else:
            self.adapter = cutadapt.adapters.BackAdapter(seq, name=name, **kw)

    def match_to(self, seq, qual):
        match = self.adapter.match_to(seq)
        n_trimmed = 0
        if match:
            if self.where == "left":
                n_trimmed = match.rstop
            else:
                n_trimmed = len(seq) - match.rstart

        return n_trimmed


class AdapterFlavor:
    def __init__(
        self,
        args,
        default_max_errors=0.1,
        default_min_overlap=3,
        default_min_base_qual=20,
        default_min_read_length=18,
        default_paired_end="single_end",
    ):
        # import cutadapt.adapters

        self.stats = defaultdict(int)
        self.total = defaultdict(int)
        self.lhist = defaultdict(int)
        self.flavor = args.flavor
        if not args.flavor in args.adapter_flavors:
            raise KeyError(
                f"adapter_flavor '{args.flavor}' not found in config.yaml! Need valid --flavor=... "
            )

        flavor_d = args.adapter_flavors[self.flavor]
        # from pprint import pprint

        # pprint(flavor_d)
        self.adapter_sequences = args.adapters

        self.trimmers_right = []
        for adap_d in flavor_d.get("cut_right", []):
            for name, param_d in adap_d.items():
                if name == "Q":
                    adapter = QualityTrim(
                        min_base_qual=param_d.get(
                            "min_base_qual", default_min_base_qual
                        )
                    )
                else:
                    adapter = AdapterTrim(
                        name=name,
                        seq=self.adapter_sequences[name],
                        where="right",
                        max_errors=param_d.get("max_errors", default_max_errors),
                        min_overlap=param_d.get("min_overlap", default_min_overlap),
                    )

                self.trimmers_right.append(adapter)

        self.trimmers_left = []
        for adap_d in flavor_d.get("cut_left", []):
            for name, param_d in adap_d.items():
                adapter = AdapterTrim(
                    name=name,
                    seq=self.adapter_sequences[name],
                    where="left",
                    max_errors=param_d.get("max_errors", default_max_errors),
                    min_overlap=param_d.get("min_overlap", default_min_overlap),
                )
                self.trimmers_left.append(adapter)

        self.min_read_length = flavor_d.get("min_read_length", default_min_read_length)
        self.paired_end = flavor_d.get("paired_end", default_paired_end)

    def process_read(self, read_seq, read_qual):
        # print(read_qual)
        start = 0
        end = len(read_seq)

        self.stats["N_input"] += 1
        self.total["bp_input"] += end
        trimmed_names_right = []
        trimmed_bases_right = []

        trimmed_names_left = []
        trimmed_bases_left = []

        def check_discard(start, end, reason):
            if (end - start) < self.min_read_length:
                self.stats[reason] += 1
                self.stats["N_discarded"] += 1
                self.total["bp_discarded"] += end - start

                return True
            else:
                return False

        for trimmer in self.trimmers_right:
            n_trimmed = trimmer.match_to(read_seq[start:end], read_qual[start:end])
            if n_trimmed:
                end -= n_trimmed
                trimmed_bases_right.append(n_trimmed)
                trimmed_names_right.append(trimmer.name)

                self.stats[f"N_{trimmer.name}_trimmed"] += 1
                self.total[f"bp_{trimmer.name}_trimmed"] += n_trimmed
                self.total[f"bp_trimmed"] += n_trimmed

            if check_discard(start, end, f"N_too_short_after_{trimmer.name}"):
                return

        for trimmer in self.trimmers_left:
            n_trimmed = trimmer.match_to(read_seq[start:end], read_qual[start:end])
            if n_trimmed:
                start += n_trimmed
                trimmed_bases_left.append(n_trimmed)
                trimmed_names_left.append(trimmer.name)

                self.stats[f"N_{trimmer.name}_trimmed"] += 1
                self.total[f"bp_{trimmer.name}_trimmed"] += n_trimmed
                self.total[f"bp_trimmed"] += n_trimmed

            if check_discard(start, end, f"N_too_short_after_{trimmer.name}"):
                return

        # we've made it to the end!
        self.stats["N_kept"] += 1
        self.total["bp_kept"] += end

        tags = []
        if trimmed_names_right:
            tags.append(f"A3:Z:{','.join(trimmed_names_right)}")
            tags.append(f"T3:Z:{','.join([str(s) for s in trimmed_bases_right])}")

        if trimmed_names_left:
            tags.append(f"A5:Z:{','.join(trimmed_names_left)}")
            tags.append(f"T5:Z:{','.join([str(s) for s in trimmed_bases_left])}")

        self.lhist[end - start] += 1
        return start, end, "\t".join(tags)


def process_reads(input, output, args):
    flavor = AdapterFlavor(args)
    # TODO: paired-end processing
    for read in input:
        cols = read.split("\t")
        read_seq = cols[9]
        qual_str = cols[10]
        # read_qual = np.array(bytearray(qual_str.encode("ASCII"))) - args.phred_base

        result = flavor.process_read(read_seq, qual_str)
        if result:
            start, end, tags = result
            cols[9] = read_seq[start:end]
            cols[10] = qual_str[start:end]

            if tags:
                cols[-1] = f"{cols[-1].rstrip()}\t{tags}\n"

            output.write("\t".join(cols))

    return {"stats": flavor.stats, "total": flavor.total, "lhist": flavor.lhist}


def main(args):
    logger = util.setup_logging(args, name="scbamtools.cutadapt_bam.main")

    w = (
        mf.Workflow("cutadapt")
        .BAM_reader(input=args.bam_in, threads=args.threads_read)
        .distribute(
            input=mf.FIFO("input_sam", "rt"),
            outputs=mf.FIFO("dist{n}", "wt", n=args.threads_work),
            chunk_size=1,
            header_detect_func=util.is_header,
            header_broadcast=False,
            header_fifo=mf.FIFO("orig_header", "wt"),
        )
        .workers(
            input=mf.FIFO("dist{n}", "rt"),
            output=mf.FIFO("out{n}", "wt"),
            func=process_reads,
            args=args,
            n=args.threads_work,
        )
        .add_job(
            func=util.update_header,
            input=mf.FIFO("orig_header", "rt"),
            output=mf.FIFO("new_header", "wt"),
            progname="trim.py",
        )
        .collect(
            inputs=mf.FIFO("out{n}", "rt", n=args.threads_work),
            header_fifo=mf.FIFO("new_header", "rt"),
            output=mf.FIFO("out_sam", "wt"),
            chunk_size=1,
        )
        .funnel(
            input=mf.FIFO("out_sam", "rt"),
            output=args.bam_out,
            _manage_fifos=False,
            func=mf.parts.bam_writer,
            threads=args.threads_write,
            fmt=f"Sh{args.bam_out_mode}",
        )
        .run()
    )
    # stats = defaultdict(int)
    # total = defaultdict(int)
    # lhist = defaultdict(int)

    # dt = time() - t0
    # logger.info(
    #     f"processed {stats['N_input']} reads in {dt:.1f} seconds ({stats['N_input']/dt:.1f} reads/second)."
    # )

    # if args.stats_out:
    #     with open(util.ensure_path(args.stats_out), "wt") as f:
    #         f.write("key\tcount\tpercent\n")
    #         for k, v in sorted(stats.items(), key=lambda x: -x[1]):
    #             f.write(f"reads\t{k}\t{v}\t{100.0 * v/stats['N_input']:.2f}\n")

    #         for k, v in sorted(total.items(), key=lambda x: -x[1]):
    #             f.write(f"bases\t{k}\t{v}\t{100.0 * v/total['bp_input']:.2f}\n")

    #         for k, v in sorted(lhist.items()):
    #             f.write(f"L_final\t{k}\t{v}\t{100.0 * v/stats['N_kept']:.2f}\n")


def cmdline():
    args = parse_args()
    util.setup_logging(args)
    main(args)


if __name__ == "__main__":
    cmdline()



__license__ = "MIT"
__authors__ = ["Marvin Jens"]
__email__ = "mjens@mit.edu"


# from cython.parallel import parallel, prange
import numpy as np
cimport numpy as np
# cimport cython
# cimport openmp

# from libc.math cimport exp, log
# from libc.stdlib cimport abort, malloc, free

ctypedef extern np.uint8_t UINT8_t
ctypedef extern np.uint16_t UINT16_t
ctypedef extern np.uint32_t UINT32_t
ctypedef extern np.uint64_t UINT64_t
ctypedef extern np.int32_t INT32_t
ctypedef extern np.float32_t FLOAT32_t
ctypedef extern np.float64_t FLOAT64_t
                                
# cdef extern UINT64_t rand_state[2]
# cdef extern UINT64_t RAND_MAX = 2**64 - 1
# cdef extern FLOAT32_t FRAND_MAX = RAND_MAX

# cdef extern UINT8_t letter_to_bits[256]



#cython: boundscheck=False, wraparound=False, initializedcheck=False, overflowcheck=False, cdivision=True, language_level=3
##cython: boundscheck=True, wraparound=True, initializedcheck=True, overflowcheck=True, cdivision=False, language_level=3
#!python

__license__ = "MIT"
__authors__ = ["Marvin Jens"]
__email__ = "marvin.jens@charite.de"

from types cimport *
from types import *

from cython.parallel import parallel, prange
import numpy as np
cimport numpy as np
cimport cython
cimport openmp

from libc.math cimport exp, log
from libc.stdlib cimport abort, malloc, free

# maps ASCII values of A,C,G,T (U) to correct bits
# N is silently converted to A
cdef UINT8_t letter_to_bits[256]
for i in range(256):
    letter_to_bits[i] = 255

letter_to_bits[ord('n')] = 0
letter_to_bits[ord('a')] = 0 
letter_to_bits[ord('c')] = 1 
letter_to_bits[ord('g')] = 2
letter_to_bits[ord('t')] = 3
letter_to_bits[ord('u')] = 3

letter_to_bits[ord('N')] = 0
letter_to_bits[ord('A')] = 0 
letter_to_bits[ord('C')] = 1 
letter_to_bits[ord('G')] = 2
letter_to_bits[ord('T')] = 3
letter_to_bits[ord('U')] = 3


cdef UINT8_t bits_to_letters[4]
bits_to_letters[:] = [ord('A'), ord('C'), ord('G'), ord('T')]
    
cpdef inline seq_to_index(str py_str):
    cdef UINT32_t L = len(py_str)
    cdef UINT8_t x = 0
    cdef UINT8_t nt = 0
    cdef bytes py_byte_str = py_str.encode('ascii')
    cdef unsigned char *seq = py_byte_str
    
    cdef UINT64_t idx = 0
    for x in range(L):
        idx = (idx << 2) | letter_to_bits[seq[x]]

    return idx

def index_to_seq(UINT64_t index, UINT8_t l):
    nucs = ['A','C','G','T']
    seq = []
    for i in range(l):
        j = index >> ((l-i-1) * 2)
        seq.append(nucs[j & 3])

    return "".join(seq)

cpdef np.ndarray[UINT64_t] hull(UINT64_t index, UINT8_t l):
    """
    Generate all 3 x l edit-distance 1 variants of the 2-bit encoded l-mer index
    """
    cdef np.ndarray[UINT64_t] variants = np.zeros(3*l + 1, dtype=np.uint64)
    cdef UINT64_t i, j, var, mask, mut, base
   
    # first entry is the unedited reference
    variants[0] = index

    for i in range(l):
        # all bits 1 except the two bits of the base we are editing
        mask = ~(<unsigned long> 3 << (i << 1))
        # NOTE: i << 1 is the same as i * 2
        for j in range(3):
            # select the bits NOT to edit
            var = index & mask
            # select the bits of the base we DO want to edit (and bring to lower bits)
            base = (index >> (i << 1)) & <unsigned long> 3
            # addition plus wrap around at 3 gives all mutations
            mut = (base + 1 + j) % <unsigned long> 4 
            # add in the edited base bits at the correct bit position again
            var |= mut << (i * 2)
            # done. store the edited l-mer
            variants[i * <unsigned long> 3 + j + 1] = var
    
    return variants

cdef packed struct bc_info_t:
    # Data structure for the barcode tree
    UINT64_t idx # can store barcodes up to 32nt length     # 8
    UINT32_t left # can store up to 4 billion barcodes    # 4
    UINT32_t right                                        # 4
    UINT16_t tile # up to 65k tiles                     # 2
    UINT16_t x # x/y positions between 0..65k           # 2
    UINT16_t y                                          # 2
    # total size = 22 bytes

ctypedef bc_info_t bc_info

cdef UINT32_t LEAF = 0xFFFFFFFE

cpdef UINT32_t read_txt_to_buf(src, bc_info_t [:] buf, UINT16_t tile=0):
    #cdef bc_info_t [::1] buf = _buf # fast memoryview into the buffer
    cdef str line
    cdef list cols
    #cdef bc_info_t[] records = malloc(sizeof(bc_info_t) * N_max)
    cdef bc_info_t* bc
    cdef UINT64_t i = 0
    for line in src:
        if line.startswith('cell_bc'):
            continue
        
        cols = line.split('\t')
        bc = &buf[i]
        bc.idx = seq_to_index(cols[0])
        bc.tile = tile
        bc.x = int(cols[1])
        bc.y = int(cols[2])
        bc.left = LEAF
        bc.right = LEAF

        # print(i, buf[i])

        i += 1
    
    return i

cpdef UINT32_t build_tree(bc_info_t [:] buf, UINT32_t low, UINT32_t high):
    cdef bc_info_t* node
    cdef UINT32_t root = 0

    if high - low < 1:
        return LEAF

    cdef UINT32_t mid = ((high - low) >> 1) + low
    node = &buf[mid]
    node.left = build_tree(buf, low, mid)
    node.right = build_tree(buf, mid + 1, high)
   
    return mid

cpdef inline UINT8_t check_contains(bc_info_t [:] buf, UINT32_t root, UINT64_t idx):
    cdef bc_info_t* node
    cdef bc_info_t* leaf
    if root == LEAF:
        return 0
    else:
        node = &buf[root]
        if node.idx == idx:
            return True
        elif node.idx > idx:
            return check_contains(buf, node.left, idx)
        elif node.idx < idx:
            return check_contains(buf, node.right, idx)

cpdef np.ndarray[UINT8_t] check_contains_batch(bc_info_t [:] buf, UINT32_t root, UINT64_t [:] indices):
    cdef UINT64_t N = len(indices)

    # output buffer
    cdef np.ndarray[UINT8_t] res_ = np.empty(N, dtype=np.uint8)
    cdef UINT8_t [::1] res = res_ # fast memory-view object
    cdef UINT64_t i
    for i in range(N):
        res[i] = check_contains(buf, root, indices[i])

    return res_


